import logging
from unittest import TestCase

from nomnomdata.engine.test import NominodeMock
from nomnomdata.engine.test_creds import credentials

from ..executable import TemplateExecutable

config = {
    "3": credentials["aws_connection"],
}
base_params = {
    "config": config,
    "aws_connection": {"connection_uuid": "3"},
    "test_parameter_1": "5",
    "test_parameter_2": False,
}


class TestReportSync(TestCase):
    def tearDown(self):
        self.logger.info("teardown")

    def setUp(self):
        self.params = base_params.copy()
        self.logger = logging.getLogger("nomigen.test")

    def test_basic(self):
        nominode = NominodeMock(task_parameters=self.params)
        with nominode:
            executable = TemplateExecutable()
            executable.do_thing()



import logging
from pprint import pformat

from cerberus.validator import Validator

from .code_sql_types import code_types, sql_types

_logger = logging.getLogger(__name__)

parameter_types = [
    "code",
    "json",
    "sql",
    "metadata_table",
    "boolean",
    "connection",
    "text",
    "int",
    "string",
    "enum",
    "enum_list",
]
help_schema = {
    "oneof_schema": [
        {"header_id": {"type": "string", "required": False}},
        {"file": {"type": "string", "required": False}},
    ],
    "type": "dict",
}


parameter_field_schema = {
    "help": help_schema.copy(),
    "collapsed": {"required": False, "type": "boolean"},
    "required": {"required": False, "type": "boolean"},
    "choices": {"dependencies": {"type": ["enum", "enum_list"]}, "type": "list"},
    "default_choices": {"dependencies": {"type": ["enum_list"]}, "type": "list"},
    "display_type": {"dependencies": {"type": ["enum_list"]}, "type": "string"},
    "connection_type_uuid": {"dependencies": {"type": ["connection"]}, "type": "string"},
    "name": {"type": "string", "required": True},
    "dialect": {
        "type": "string",
        "oneof": [
            # If type is "sql", only check the list of sql types.
            # If type is "code", only check the list of language types (excluding sql).
            {"dependencies": {"type": ["sql"]}, "allowed": sql_types},
            {"dependencies": {"type": ["code"]}, "allowed": code_types},
        ],
    },
    "description": {"type": "string", "required": False},
    "display_name": {"type": "string", "required": True},
    "type": {
        "allowed": parameter_types,
        "required": True,
        "oneof": [
            # Bi-lateral dependencies
            # E.G. a field that is a "connection" type must have "connection_type_uuid" as a parameter.
            {"allowed": ["connection"], "dependencies": "connection_type_uuid"},
            {"allowed": ["code", "sql"], "dependencies": "dialect"},
            {"allowed": ["enum", "enum_list"], "dependencies": "choices"},
            # Add above bi-lateral dependencies as "forbidden" so above dependencies are evaluated only once.
            {"forbidden": ["connection", "code", "sql", "enum", "enum_list"]},
        ],
    },
    "shared_object_type_uuid": {"type": "string", "required": False},
    "max": {"required": False, "type": "integer"},
    "min": {"required": False, "type": "integer"},
    "default": {"required": False},
}
parameter_group_schema = {
    "display_name": {"type": "string", "required": True},
    "collapsed": {"required": False, "type": "boolean",},
    "shared_parameter_group_uuid": {
        "required": False,
        "dependencies": {"type": ["group"]},
        "type": "string",
    },
    "description": {"type": "string", "required": False},
    "name": {"type": "string", "required": True},
    "type": {"allowed": ["group"], "required": True},
    "parameters": {
        "type": "list",
        "schema": {"type": "dict", "schema": parameter_field_schema.copy()},
    },
}


engine_schema = {
    "uuid": {
        "type": "string",
        "minlength": 8,
        "required": True,
        "regex": r"^[A-Z0-9]+([_-][A-Z0-9]+)*$",
    },
    "icons": {
        "type": "dict",
        "schema": {
            "1x": {"type": "string"},
            "2x": {"type": "string"},
            "3x": {"type": "string"},
        },
    },
    "help": help_schema.copy(),
    "alias": {"type": "string", "required": True},
    "description": {"type": "string", "required": True},
    "nnd_model_version": {"type": "integer", "required": True},
    "options": {"required": False, "type": "dict"},
    "location": {
        "type": "dict",
        "schema": {
            "repo": {"type": "string", "required": True},
            "image": {"type": "string", "required": True},
            "region": {"type": "string", "required": True},
            "repo_type": {"type": "string", "required": True},
        },
    },
    "categories": {
        "type": "list",
        "required": True,
        "schema": {
            "type": "dict",
            "schema": {"name": {"type": "string", "required": True}},
        },
    },
    "type": {"allowed": ["engine"], "required": True},
    "actions": {
        "type": "dict",
        "required": True,
        "keyschema": {"type": "string"},
        "valueschema": {
            "type": "dict",
            "schema": {
                "description": {"type": "string", "required": True},
                "help": help_schema.copy(),
                "display_name": {"type": "string", "required": True},
                "parameters": {
                    "type": "list",
                    "required": True,
                    "schema": {"type": "dict", "schema": parameter_group_schema},
                },
            },
        },
    },
}


shared_object_type_schema = {
    "uuid": {
        "type": "string",
        "minlength": 8,
        "required": True,
        "regex": r"^[A-Z\d]+-[A-Z\d]+$",
    },
    "alias": {"type": "string", "required": True},
    "description": {"type": "string", "required": True},
    "nnd_model_version": {"type": "integer", "required": True},
    "categories": {
        "type": "list",
        "required": True,
        "schema": {
            "type": "dict",
            "schema": {"name": {"type": "string", "required": True}},
        },
    },
    "type": {"allowed": ["shared_object_type"], "required": True},
    "parameters": {
        "type": "list",
        "schema": {
            "type": "dict",
            "anyof_schema": [
                parameter_field_schema.copy(),
                parameter_group_schema.copy(),
            ],
        },
    },
}
connection_type_schema = {
    "uuid": {
        "type": "string",
        "minlength": 8,
        "required": True,
        "regex": r"^[A-Z\d]+-[A-Z\d]+$",
    },
    "alias": {"type": "string", "required": True},
    "description": {"type": "string", "required": True},
    "nnd_model_version": {"type": "integer", "required": True},
    "categories": {
        "type": "list",
        "required": True,
        "schema": {
            "type": "dict",
            "schema": {"name": {"type": "string", "required": True}},
        },
    },
    "type": {"allowed": ["connection"], "required": True},
    "parameters": {
        "type": "list",
        "schema": {"type": "dict", "schema": parameter_group_schema.copy()},
    },
}
schemas = {
    "engine": engine_schema,
    "shared_object_type": shared_object_type_schema,
    "connection": connection_type_schema,
}


def validate_model(model):
    schema = schemas[model["type"]]
    validator = Validator(schema)
    if not validator.validate(model):
        _logger.error(
            f"Errors validating {model['type']} model \n\t\t{pformat(validator.errors)}"
        )
        return False
    else:
        return True



import shutil
from os import makedirs, path

import click
from jinja2 import Template


@click.command()
@click.argument("engine_name", required=True)
def create_new(engine_name):
    """Create a template engine layout based on the given name"""
    t_dir = path.join(path.dirname(__file__), "template")

    def copymap(src, dest):
        if path.basename(src).startswith("ex_"):
            dest = path.join(path.dirname(dest), path.basename(src)[3:])
        with open(src, "r") as infile:
            template = Template(infile.read())
        makedirs(path.dirname(dest), exist_ok=True)
        template = template.render(engine_name=engine_name)
        with open(dest, "w") as outfile:
            outfile.write(template)

    shutil.copytree(t_dir, engine_name, copy_function=copymap)



import click

from . import __version__
from .create_new import create_new
from .deploy import deploy
from .model_update import model_update


@click.group(name="engine-tools")
@click.version_option(version=__version__, prog_name="nomnomdata-engine-tools")
def cli():
    """Used for building/deploying engines"""


@cli.command()
def test():
    """Run unittests for nomnomdata engine tools"""
    import pytest

    pytest.main(["--pyargs", "nomnomdata.tools.engine"])


cli.add_command(create_new)
cli.add_command(deploy)
cli.add_command(model_update)



code_types = [
    "abap",
    "abc",
    "actionscript",
    "ada",
    "apache_conf",
    "apex",
    "applescript",
    "asciidoc",
    "asl",
    "assembly_x86",
    "autohotkey",
    "batchfile",
    "bro",
    "c9search",
    "c_cpp",
    "cirru",
    "clojure",
    "cobol",
    "coffee",
    "coldfusion",
    "csharp",
    "csound_document",
    "csound_orchestra",
    "csound_score",
    "csp",
    "css",
    "curly",
    "d",
    "dart",
    "diff",
    "django",
    "dockerfile",
    "dot",
    "drools",
    "edifact",
    "eiffel",
    "ejs",
    "elixir",
    "elm",
    "erlang",
    "forth",
    "fortran",
    "fsharp",
    "fsl",
    "ftl",
    "gcode",
    "gherkin",
    "gitignore",
    "glsl",
    "gobstones",
    "golang",
    "graphqlschema",
    "groovy",
    "haml",
    "handlebars",
    "haskell",
    "haskell_cabal",
    "haxe",
    "hjson",
    "html",
    "html_elixir",
    "html_ruby",
    "ini",
    "io",
    "jack",
    "jade",
    "java",
    "javascript",
    "json",
    "jsoniq",
    "jsp",
    "jssm",
    "jsx",
    "julia",
    "kotlin",
    "latex",
    "less",
    "liquid",
    "lisp",
    "livescript",
    "logiql",
    "logtalk",
    "lsl",
    "lua",
    "luapage",
    "lucene",
    "makefile",
    "markdown",
    "mask",
    "matlab",
    "maze",
    "mel",
    "mixal",
    "mushcode",
    "nix",
    "nsis",
    "objectivec",
    "ocaml",
    "pascal",
    "perl",
    "perl6",
    "php",
    "php_laravel_blade",
    "pig",
    "plain_text",
    "powershell",
    "praat",
    "prolog",
    "properties",
    "protobuf",
    "puppet",
    "python",
    "r",
    "razor",
    "rdoc",
    "red",
    "rhtml",
    "rst",
    "ruby",
    "rust",
    "sass",
    "scad",
    "scala",
    "scheme",
    "scss",
    "sh",
    "sjs",
    "slim",
    "smarty",
    "snippets",
    "soy_template",
    "space",
    "sparql",
    "stylus",
    "svg",
    "swift",
    "tcl",
    "terraform",
    "tex",
    "text",
    "textile",
    "toml",
    "tsx",
    "turtle",
    "twig",
    "typescript",
    "vala",
    "vbscript",
    "velocity",
    "verilog",
    "vhdl",
    "visualforce",
    "wollok",
    "xml",
    "xquery",
    "yaml",
]


sql_types = [
    "pgsql",
    "mysql",
    "redshift",
    "sql",
    "sqlserver",
]



from nomnomdata.engine.api import NominodeClient
from nomnomdata.engine.core import Executable

__author__ = "Nom Nom Data LLC"
__copyright__ = "Copyright 2018, Nom Nom Data LLC"


class TemplateExecutable(Executable):
    def __init__(self):
        super().__init__("nomigen.{{engine_name}}")
        self.creds = {}
        self.creds["aws_connection"] = self.params["config"][
            self.params["aws_connection"]["connection_uuid"]
        ]
        self.nominode = NominodeClient()

    def do_thing(self):
        self.logger.info("I did a thing!")
        self.nominode.update_progress(message="I did it!")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nomnomdata',
 'nomnomdata.tools.engine',
 'nomnomdata.tools.engine.template.engine.pkg',
 'nomnomdata.tools.engine.template.engine.pkg.tests',
 'nomnomdata.tools.engine.tests']

package_data = \
{'': ['*'], 'nomnomdata.tools.engine': ['template/*', 'template/engine/*']}

install_requires = \
['boto3>=1.12.47,<2.0.0',
 'cerberus>=1.3.2,<2.0.0',
 'docker-compose>=1.25.5,<2.0.0',
 'docker[ssh]>=4.2.0,<5.0.0',
 'dunamai>=1.1.0,<2.0.0',
 'fsspec>=0.7.3,<0.8.0',
 'gitpython>=3.1.1,<4.0.0',
 'jinja2>=2.11.2,<3.0.0',
 'nomnomdata-auth>=2.2.0,<3.0.0',
 'nomnomdata-cli>=0.1.0,<0.2.0',
 'python-dotenv>=0.13.0,<0.14.0',
 'pyyaml>=5.3.1,<6.0.0',
 'requests>=2.23.0,<3.0.0',
 's3fs>=0.4.2,<0.5.0',
 'sqlparse>=0.3.1,<0.4.0']

entry_points = \
{'nomnomdata.cli_plugins': ['engine-tools = nomnomdata.tools.engine.cli:cli']}

setup_kwargs = {
    'name': 'nomnomdata-tools-engine',
    'version': '0.8.0.post2.dev0',
    'description': 'Package containing tooling for developing nominode engines',
    'long_description': '# Installation\n\n`pip install nomnomdata-engine-tools nomnomdata-engine`\n\n# Creating a new engine\n\nThis will walk through all the information required to create, build and deploy a new engine.\n\nDirectory Structure\n\nInitialize a new engine with\n\n`nnd engine-tools create-new <ENGINE_NAME>`\n\nThis will create a new directory with the layout\n\n``` directory\n.\n├── .dockerignore\n├── .gitignore\n├── .ex_pyproject.toml\n└── engine\n    ├── build.dockerfile\n    ├── model.yaml\n    ├── requirements.txt\n    └── pkg\n        ├── __init__.py\n        ├── executable.py\n        └── tests\n            ├── __init__.py\n            └── test_executable.py\n```\n\n## build.dockerfile\n\nEngines are deployed as docker images, so for every engine you must have a build.dockerfile located next to a model.yaml file. You will want to place your python files the \'pkg\' directory if you want nnd engine-tools build -rt to work properly.\n\nLook at the template build.dockerfile for an example and more information. You must include\n\n`CMD nnd engine run pkg/executable.py`\n\n Or your engine will not start correctly.\n\n## requirements.txt\n\nSimple pip requirements file. Requirements should be set to a specific version (ie requests==2.18.4) . Only requirements specific to your engine should be included in here. The template requirements.txt shows how to include the packages nomnomdata-cli & nomnomdata-engine which are the easiest way to run your executable.\n\n## executable.py\n\nThis is where the code that actually gets run lives.\n\n``` python\nfrom nomnomdata.engine.api import NominodeClient\nfrom nomnomdata.engine.core import Executable\n\nclass TemplateExecutable(Executable):\n    def __init__(self):\n        # Very important to call the inherited __init__ before anything else\n        super().__init__("nomigen.{{engine_name}}")\n\n    def do_thing(self):\n        self.logger.info("I did a thing!")\n        self.logger.info(f"My parameters are {self.params}")\n        self.nominode.update_progress(message="I did it!")\n\n```\n\n## test_executable.py\n\nTests! Our run tests command uses pytest to run these inside your docker image, so you can use anything pytest supports. The most important thing is to wrap your test code with a NominodeMock , this essentially sets up a \'fake\' nomnominode catching http requests to it.\n\n```python\nfrom nomnomdata.engine.test import NominodeMock\nfrom nomnomdata.engine.test_creds import credentials\nfrom ..executable import TemplateExecutable\n\nconfig = {\n\n    1: credentials["aws_connection"],\n\n}\n\nparams = {\n    "config": config,\n    \'action_name\': \'test\',\n    "aws_connection": {\n        "connection_uuid": \'2\'\n    },\n    \'other_param\': \'some variable\'\n\n}\n\n# the most basic test possible..\ndef test_init(self):\n    with NominodeMock(params):\n        t = TemplateExecutable()\n        t.do_thing()\n\n```\n\n## model.yaml\n\nThis is the description of your engine that will reside in the nomitall database, used by the nominode_ui to dynamically create configuration pages.\n\n```yaml\n# the unique identifier for this engine, must be unique across all engines\nuuid: NOMNOM_EXAMPLE_DOWNLOADER\nalias: "EXAMPLE: Downloader"\ndescription: An Example\n# repo information, most of the information is static apart from image\nlocation:\n  repo: 445607516549.dkr.ecr.us-east-1.amazonaws.com\n  # the :testing tag at the end indicates which ECS tag to use, for staging this should be testing, prod should use alpha\n  image: example/api/downloader:testing\n  region: us-east-1\n  repo_type: aws\n# categories this engine fits into, there are no fixed categories at this time\ncategories:\n\n  + name: example\n  + name: s3\n  + name: nomnomdata\n\n# this is a list of actions this engine can perform, these map directly to\n# functions of your Executable class in executable.py\nparameters:\n\n  + action_name: example_function\n\n    description: An example\n    display_name: Example Function\n    # this is where you can add what connections the engine requires\n    # also what adhoc parameters the engine uses.\n    # these will be presented to the user to fill in\n    parameters:\n    # this include parameter is a special directive that allows you include parameters from other files\n    # includes that are specific to the engine should be stored in your_engine/models\n\n    - include:\n      - \'common/meta_types/user_range.yaml\'\n    - connection_type_uuid: APP1E-T0NXM\n\n      parameter_name: app_annie_token\n      display_name: App Annie Token\n      description:  Credential token used to access the App Annie api.\n      required: true\n      type: connection\n\n    - parameter_name: s3_info\n\n      display_name: S3 Info\n      type: group\n\n    - connection_type_uuid: AWS5D-TO99M\n\n      description: AWS Credentials with access to the s3 bucket where the data table lives.\n      display_name: AWS Token Credentials\n      parameter_name: aws_connection\n      parameter_group: s3_info\n      required: false\n      type: connection\n\n    - connection_type_uuid: AWSS3-BUCKT\n\n      description: S3 Bucket where you want to unload the data to.\n      display_name: S3 Bucket\n      parameter_name: s3_bucket\n      parameter_group: s3_info\n      required: false\n      type: connection\n\n    - description: Path to a location where data files will be exported\n\n      display_name: S3 Path\n      parameter_name: s3_path\n      parameter_group: s3_info\n      required: True\n      type: string\n      max: 2048\n\n    - parameter_name: format_options\n\n      display_name: File/Formatting Options\n      type: group\n\n    - description: If specified will insert the date into the folder structure.  Example... y=1998/m=12/d=25/file_name.json.gz.\n\n      display_name: Date Folders\n      parameter_name: append_date\n      parameter_group: format_options\n      required: True\n      type: enum\n      default: True\n      choices:\n\n        - True\n        - False\n\n```\n\n## nomigen_test_credentials.json\n\nThis file exists outside the engine folders, and should be placed in your copy of the engines-config repo, you can find an example of it there. This is where you can store your credentials required by tests. If you set the environment variable NOMIGEN_TEST_CREDENTIALS to the location of your copy of engines-config, these creds will be made accessible via the nomnomdata.engine.test_credentials module at runtime.\n\n``` JSON\n{\n    "app_annie_token": {\n        "token": "your_app_annie_token"\n    },\n    "aws_connection": {\n        "aws_access_key_id": "your_aws_access_key",\n        "aws_secret_access_key": "your_aws_secret_access_key"\n    },\n    "s3_bucket": {\n        "bucket": "shughes-test-bucket"\n    }\n}\n```\n',
    'author': 'Nom Nom Data Inc',
    'author_email': 'info@nomnomdata.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/nomnomdata/tools/nomnomdata-tools-engine',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import functools
import logging
import sys
from os import getcwd, path
from pathlib import Path
from pprint import pformat
from subprocess import check_output
from urllib.parse import urljoin, urlparse

import click
import fsspec
import requests
import yaml
from git import Repo

from nomnomdata.auth import DEFAULT_PROFILE, NNDAuth, get_profiles

from .model_validator import validate_model

_logger = logging.getLogger(__name__)


class NomitallSession(requests.Session):
    def __init__(self, prefix_url=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.prefix_url = prefix_url

    def request(self, method, url, *args, **kwargs):
        url = urljoin(self.prefix_url, url)
        _logger.info(f"Request {method}:{url}")
        return super().request(method, url, *args, **kwargs)


def confirmation_prompt(question):
    reply = input(question + " (y or n): ")
    if reply[0] == "y":
        return True
    if reply[0] == "n":
        return False
    else:
        return confirmation_prompt("Please enter ")


@functools.lru_cache()
def fetch_remote_git_yaml(remote_url, git_ref, file_path):
    cmd = "git archive --remote={} {} {} | tar -xO".format(remote_url, git_ref, file_path)
    file_str = check_output(cmd, shell=True).decode()
    return yaml.full_load(file_str)


def fetch_git_yaml(relpath, git_ref):
    if git_ref:
        repo = Repo(".", search_parent_directories=True)
        reldir = path.relpath(getcwd(), repo.working_dir)
        git_path = path.join(reldir, relpath)
        _logger.info("\tUsing {} at ref {}, will fetch latest".format(git_path, git_ref))
        for fetch_info in repo.remotes["origin"].fetch():
            _logger.info("\tUpdated %s to %s" % (fetch_info.ref, fetch_info.commit))
        model_str = repo.git.show("{}:{}".format(git_ref, git_path))
        return yaml.full_load(model_str)
    else:
        with open(relpath, "r") as f:
            _logger.info("Using {} from current working directory".format(relpath))
            return yaml.full_load(f)


def confirm_git_branch(work_dir, nomitall):
    if work_dir:
        if nomitall == "nomitall-prod":
            question = "You are about to push your working directory to production!\
            Are you absolutely sure you want to do this?"

            if not confirmation_prompt(question):
                _logger.info("You have chosen wisely")
                sys.exit()
            if not confirmation_prompt("Absolutely sure?"):
                _logger.info("You have chosen wisely")
                sys.exit()
        branch = None
    else:
        if nomitall == "nomitall-prod":
            branch = "origin/master"
        elif nomitall == "nomitall-stage":
            branch = "origin/staging"
    return branch


def fetch_include(fpath, git_ref, model_fp, remote=None):
    if remote:
        return fetch_remote_git_yaml(remote, git_ref, fpath)
    else:
        final_path = Path(path.abspath(model_fp)).parent / Path(fpath)
        return fetch_git_yaml(final_path.resolve(), git_ref)


def include_includes(parameters, git_ref, model_fp, remote=None):
    parsed_params = []
    assert isinstance(
        parameters, list
    ), "Parameters lists must be a list of dictionaries ( `- foo:bar` , not `foo:bar` )"
    for parameter in parameters:
        assert isinstance(
            parameter, dict
        ), "Parameters lists must be a list of dictionaries ( `- foo:bar` , not `foo:bar` )"
        if "parameters" in parameter:
            parameter["parameters"] = include_includes(
                parameter["parameters"], git_ref, model_fp
            )
            parsed_params.append(parameter)
        elif "include" in parameter:
            for include in parameter["include"]:
                _logger.info(
                    "\tIncluding file {} from {}@{}".format(include, remote, git_ref)
                )
                include_doc = fetch_include(include, git_ref, model_fp, remote)
                _logger.debug(f"FETCHED INCLUDE: \n{pformat(include_doc)}")
                _logger.debug("PARSING FETCHED FOR INCLUDES")
                include_doc = include_includes(include_doc, git_ref, model_fp, remote)
                _logger.debug(
                    f"FETCHED INCLUDE AFTER RESCURSIVE INCLUDE: \n{pformat(include_doc)}"
                )
                parsed_params.extend(include_doc)
        elif "include_git" in parameter:
            for include in parameter["include_git"]:
                new_remote, new_ref, fpath = include.split(" ")
                _logger.info(
                    "\tIncluding file {} from {}@{}".format(fpath, new_remote, new_ref)
                )
                include_doc = fetch_include(fpath, new_ref, model_fp, new_remote)
                _logger.debug(f"FETCHED INCLUDE: \n{pformat(include_doc)}")
                include_doc = include_includes(include_doc, new_ref, model_fp, new_remote)
                _logger.debug(
                    f"FETCHED INCLUDE AFTER RESCURSIVE INCLUDE: \n{pformat(include_doc)}"
                )
                parsed_params.extend(include_doc)
        else:
            parsed_params.append(parameter)
    return parsed_params


@click.command()
@click.option(
    "-wd",
    "--use_work_dir",
    "--using-work-dir",
    "work_dir",
    is_flag=True,
    help="Use model files from current working directory instead of git",
)
@click.option(
    "-n",
    "--nomitall",
    default="nomitall-stage",
    help="Specify the nomitall to update [nomitall-prod,nomitall-stage,custom_url]",
)
@click.option(
    "-dr",
    "--dry_run",
    "--dry-run",
    is_flag=True,
    help="Do not update nomitall, just output parsed model json",
)
@click.option(
    "-f",
    "--file",
    "model_fn",
    default="model.yaml",
    help="Model YAML file to build + deploy",
)
@click.option(
    "-o", "--org", "org", help="UUID of the organization to publish the model as",
)
def model_update(
    work_dir=None,
    nomitall_secret=None,
    nomitall=None,
    dry_run=None,
    model_fn=None,
    org=None,
):
    """Update staging or prod nomitall model definitions. Defaults to using files from git master/staging branch for prod/staging"""
    if nomitall == "nomitall-prod":
        nomitall_url = "https://user.api.nomnomdata.com/api/1/"
    elif nomitall == "nomitall-stage":
        nomitall_url = "https://staging.user.api.nomnomdata.com/api/1/"
    else:
        nomitall_url = nomitall
    if not org:
        profile = get_profiles()[DEFAULT_PROFILE]
        netloc = urlparse(nomitall_url).netloc
        org = profile[netloc]["default-org"]
    relpath = model_fn
    branch = confirm_git_branch(work_dir, nomitall)
    doc = fetch_git_yaml(relpath, branch)
    model_type = doc.get("type")
    doc.pop("docker", None)
    if not model_type:
        _logger.error("Model does not have a type, this is required")
        sys.exit(1)
    # add legacy model verison if it doesn't exit
    if model_type == "engine":
        for action, action_dict in doc["actions"].items():
            _logger.info("Parsing {}".format(action))
            action_dict["parameters"] = include_includes(
                action_dict["parameters"], branch, model_fn
            )
        try:
            image, tag = doc["location"]["image"].split(":")
        except ValueError:
            image = doc["location"]["image"]
        if nomitall == "nomitall-prod":
            tag = "prod"
        elif nomitall == "nomitall-stage":
            tag = "stage"
        else:
            tag = "stage"
        doc["location"]["image"] = ":".join([image, tag])
    else:
        doc["parameters"] = include_includes(doc["parameters"], branch, model_fn)

    if not validate_model(doc):
        raise click.Abort("Model did not pass validation")

    if dry_run:
        _logger.info("PARSED MODEL :")
        _logger.info(pformat(doc))
        sys.exit()

    update_model(nomitall_url, nomitall_secret, doc, model_type, org)


def upload_icons(icons, uuid, session):
    _logger.info("Pushing icons to nomitall")
    icon_files = {}
    for size, icon_uri in icons.items():
        openfile = fsspec.open(icon_uri, mode="rb")
        with openfile as f:
            icon_files[size] = f.read()
    resp = session.request("POST", f"engine/upload-icons/{uuid}", files=icon_files)
    check_response(resp)


def upload_help(help_files, uuid, session):
    _logger.info("Pushing mds to nomitall")
    resp = session.request("POST", f"engine/upload-md/{uuid}", files=help_files)
    check_response(resp)


def load_help_file(md_path):
    _logger.info(f"Loading {md_path}")
    openfile = fsspec.open(md_path, mode="rb")
    with openfile as f:
        md_file = f.read()
    return md_file


def process_help(doc, key="root"):
    help_files = {}
    if isinstance(doc, list):
        for subele in doc:
            help_files.update(process_help(subele, key=key))
    if isinstance(doc, dict):
        process_help
        for k, subele in doc.items():
            if k == "help":
                if "file" in subele:
                    if "name" in doc:
                        key += "." + doc["name"]
                    help_files[key] = load_help_file(subele["file"])
                    doc[k] = {"key": key}
            else:
                help_files.update(process_help(subele, key=".".join([key, k]),))
    return help_files


def update_model(nomitall_url, nomitall_secret, model, model_type, org):
    session = NomitallSession(prefix_url=nomitall_url)
    session.auth = NNDAuth()
    help_files = {}
    icons = {}
    if model_type == "engine":
        help_files = process_help(model)
        icons = model.pop("icons", {})
        uuid = model["uuid"]
        _logger.info("Pushing engine to nomitall")
        resp = session.request("POST", f"engine/deploy/{org}", json=model)
        check_response(resp)
        upload_help(help_files=help_files, uuid=uuid, session=session)
        upload_icons(icons=icons, uuid=uuid, session=session)

    elif model_type == "shared_object_type":
        _logger.info("Pushing shared object type to nomitall")
        resp = session.request("POST", "shared_object_type/update", json=model)
        check_response(resp)

    elif model_type == "connection":
        _logger.info("Pushing connection type to nomitall")
        resp = session.request("POST", "connection_type/update", json=model)
        check_response(resp)


def check_response(resp):
    if not resp.ok:
        if resp.status_code == 500:
            _logger.error(f"Internal server error\n{resp.text}")
            raise click.Abort
        reply_data = resp.json()
        if reply_data.get("error"):
            raise click.Abort(str(reply_data))
        if reply_data.get("status"):
            if reply_data["status"] == "success":
                _logger.info("Request successful")
        if resp.status_code == 401:
            _logger.error(f"Check secret key is valid\n\t\t {resp.json()}")
            raise click.Abort
        if resp.status_code == 403:
            _logger.error(f"Check user permissions\n\t\t {resp.json()}")
            raise click.Abort
    resp.raise_for_status()



import logging
import sys
from os import chdir, path
from urllib.parse import urljoin

import click
import docker
import requests
import yaml
from compose.progress_stream import stream_output

from nomnomdata.auth import NNDAuth

_logger = logging.getLogger(__name__)


class NomitallSession(requests.Session):
    def __init__(self, prefix_url=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.prefix_url = prefix_url

    def request(self, method, url, *args, **kwargs):
        url = urljoin(self.prefix_url, url)
        _logger.info(f"Request {method}:{url}")
        return super().request(method, url, *args, **kwargs)


@click.command(name="deploy")
@click.option(
    "-p",
    "--path",
    "buildp",
    default=".",
    show_default=True,
    help="Docker build context path",
)
@click.option(
    "-f",
    "--model-file",
    "mpath",
    default="./",
    help="Path to the engine model.yaml to build, if this is a folder model.yaml will be assumed to be in there "
    "Examples: nomnom/test/waittask/model.yaml",
)
@click.option(
    "-d",
    "--docker-file",
    "dpath",
    default="./",
    help="Path to the engine dockerfile to build, if this is a folder Dockerfile will be assumed to be in there "
    "Examples: nomnom/test/waittask/engine.dockerfile",
)
@click.option(
    "-n",
    "--nomitall",
    default="nomitall-stage",
    help="Specify the nomitall to update [nomitall-prod,nomitall-stage,custom_url]",
)
@click.option("--dry-run", is_flag=True, help="Build engine but do not deploy")
@click.pass_context
def deploy(ctx, buildp, mpath, dpath, nomitall, dry_run):
    "Build engine docker images and optionally run tests."
    if nomitall == "nomitall-prod":
        nomitall_url = "https://user.api.nomnomdata.com/api/1/"
        tag = "prod"
    elif nomitall == "nomitall-stage":
        nomitall_url = "https://staging.user.api.nomnomdata.com/api/1/"
        tag = "stage"
    else:
        nomitall_url = nomitall
        tag = "stage"
    chdir(buildp)
    if path.isdir(mpath):
        mpath = path.join(mpath, "model.yaml")
    if path.isdir(dpath):
        dpath = path.join(dpath, "Dockerfile")
    engine_model = yaml.full_load(open(mpath))
    engine_uuid = engine_model["uuid"]

    session = NomitallSession(prefix_url=nomitall_url)
    session.auth = NNDAuth()
    resp = session.request("get", f"engine/create-deploy-token/{engine_uuid}")
    data = resp.json()
    engine_name = data["repoUrl"]
    client = docker.from_env()
    try:
        client.ping()
    except (requests.ConnectionError, docker.errors.APIError) as e:
        raise Exception(
            "There was a problem connecting to the docker agent, ensure it is running and in a good state"
        ) from e
    image = build_engine(
        engine_name=engine_name, client=client, tag=tag, docker_file=dpath,
    )
    if not dry_run:
        deploy_engine(image, data)


def deploy_engine(image, creds):
    _logger.info("----Pushing image----")
    client = docker.from_env()
    client.login(
        username=creds["user"],
        password=creds["token"],
        registry=creds["registryUrl"],
        reauth=True,
    )
    push_logs = client.api.push(image, stream=True)
    list(stream_output(push_logs, sys.stdout))
    _logger.info("----Successfully pushed image----")


def build_engine(engine_name, client, tag, docker_file, buildargs=None):
    _logger.info("----Building image----")
    build_logs = client.api.build(
        dockerfile=docker_file,
        tag=":".join([engine_name, tag]),
        rm=True,
        path=".",
        pull=True,
        buildargs=buildargs,
    )
    list(stream_output(build_logs, sys.stdout))
    _logger.info("----Finished building image----")
    return ":".join([engine_name, tag])
